//! Read the temperature at regular intervals and publish that data via MQTT:
//! Configure the microcontroller with commands sent across an usb-serial connection.

#![no_std]
#![no_main]

use core::str::FromStr;

use byte_handler::cmd_task;
use cyw43::Control;
use defmt::{error, info};
use embassy_executor::Spawner;
use embassy_net::Stack;
use embassy_time::Timer;
use flash::{Config, FlashError};
use mqtt_protocol::publish::MAX_TOPIC_LENGTH;
use static_cell::StaticCell;
use temperature::{temperature_task, Temperature};
use usb_serial::{init_usb, usb_task};
use wifi::connect;
use {defmt_rtt as _, panic_probe as _};

use crate::{mqtt::BUFFER_SIZE, wifi::init_wifi};

mod byte_handler;
mod cmd_handlers;
mod flash;
mod mqtt;
mod temperature;
mod usb_serial;
mod wifi;

async fn connect_wifi<'a>(
    config: &mut Config<'static>,
    wifi: &mut Control<'a>,
    net_stack: &'a Stack<cyw43::NetDriver<'a>>,
) -> Result<(), FlashError> {
    let mut ssid_buffer = [0; flash::ENTRY_SIZE as usize];
    let ssid = config.read_config(flash::SSID_KEY, &mut ssid_buffer)?;
    let mut ssid_pw_buffer = [0; flash::ENTRY_SIZE as usize];
    let ssid_pw = config.read_config(flash::SSID_PW_KEY, &mut ssid_pw_buffer)?;

    if !ssid.is_empty() {
        connect(wifi, net_stack, ssid, ssid_pw).await;
    }
    Ok(())
}

static CONFIG: StaticCell<Config> = StaticCell::new();

static RX_BUFFER: StaticCell<[u8; BUFFER_SIZE]> = StaticCell::new();
static TX_BUFFER: StaticCell<[u8; BUFFER_SIZE]> = StaticCell::new();

static CLIENT_ID_BUFFER: StaticCell<[u8; flash::ENTRY_SIZE as usize]> = StaticCell::new();
static MQTT_IP_BUFFER: StaticCell<[u8; flash::ENTRY_SIZE as usize]> = StaticCell::new();

#[embassy_executor::main]
async fn main(spawner: Spawner) {
    let p = embassy_rp::init(Default::default());

    Timer::after_millis(100).await;

    let config = CONFIG.init(Config::init(p.FLASH, p.DMA_CH0));

    let (mut wifi, net_stack) = init_wifi(
        spawner, p.PIN_23, p.PIN_24, p.PIN_25, p.PIN_29, p.PIO0, p.DMA_CH1,
    )
    .await;
    match connect_wifi(config, &mut wifi, net_stack).await {
        Ok(_) => info!("Connected to WiFi."),
        Err(_) => error!("Failed to connect to WiFi"),
    }

    let rx_buffer = RX_BUFFER.init([0u8; BUFFER_SIZE]);
    let tx_buffer = TX_BUFFER.init([0u8; BUFFER_SIZE]);

    let client_id_buffer = CLIENT_ID_BUFFER.init([0u8; flash::ENTRY_SIZE as usize]);
    let client_id = config
        .read_config(flash::CLIENT_ID_KEY, client_id_buffer)
        .unwrap_or("tomato");

    let mqtt_ip_buffer = MQTT_IP_BUFFER.init([0u8; flash::ENTRY_SIZE as usize]);
    let mqtt_ip = config
        .read_config(flash::MQTT_KEY, mqtt_ip_buffer)
        .unwrap_or("127.0.0.1");

    let mqtt = mqtt::Client::new(client_id, net_stack, rx_buffer, tx_buffer);
    let host_addr = embassy_net::IpAddress::from_str(mqtt_ip)
        .unwrap_or(embassy_net::IpAddress::v4(127, 0, 1, 1));
    let temps = Temperature::new(p.ADC, p.ADC_TEMP_SENSOR);

    let (usb, class) = init_usb(p.USB);
    spawner
        .spawn(usb_task(usb))
        .expect("failed to run usb task");
    spawner
        .spawn(cmd_task(class, config))
        .expect("failed initializing usb command handler");
    spawner
        .spawn(temperature_task(client_id, host_addr, mqtt, temps))
        .expect("failed starting temperature task");
}
