use defmt::{error, info};
use embassy_net::IpAddress;
use embassy_rp::{
    adc::{Adc, Async, Channel, Config, Error, InterruptHandler},
    bind_interrupts,
    peripherals::{ADC, ADC_TEMP_SENSOR},
};
use embassy_time::Timer;
use mqtt_protocol::publish::MAX_TOPIC_LENGTH;

use crate::mqtt;

bind_interrupts!(struct Irqs {
    ADC_IRQ_FIFO => InterruptHandler;
});

#[embassy_executor::task]
pub async fn temperature_task(
    client_id: &'static str,
    host_addr: IpAddress,
    mut mqtt: mqtt::Client<'static>,
    mut temps: Temperature<'static>,
) -> ! {
    const TOPIC_PREFIX: &str = "temps/";
    let mut topic_buffer = [0u8; MAX_TOPIC_LENGTH - TOPIC_PREFIX.len()];
    topic_buffer[..TOPIC_PREFIX.len()].copy_from_slice(TOPIC_PREFIX.as_bytes());

    let client_id_bytes = client_id.as_bytes();
    topic_buffer[TOPIC_PREFIX.len()..TOPIC_PREFIX.len() + client_id_bytes.len()]
        .copy_from_slice(client_id_bytes);

    let topic_full =
        core::str::from_utf8(&topic_buffer[..TOPIC_PREFIX.len() + client_id_bytes.len()])
            .unwrap_or("temps/tomato");

    loop {
        match temps.read().await {
            Ok(temp) => {
                info!("publishing temperature data...");
                if let Err(e) = mqtt
                    .publish(host_addr, topic_full, &temp.to_le_bytes())
                    .await
                {
                    error!("failed to publish temperature data: {}", e);
                }
            }
            Err(e) => error!("failed to read temperature: {}", e),
        }

        Timer::after_secs(10).await;
    }
}

/// Convert the raw temperature sensor reading into degrees celsius.
///
/// Taken from the [embassy examples](https://github.com/embassy-rs/embassy/blob/b6fc682117a41e8e63a9632e06da5a17f46d9ab0/examples/rp/src/bin/adc.rs#L43).
fn convert_to_celsius(raw_temp: u16) -> f32 {
    // According to chapter 4.9.5. Temperature Sensor in RP2040 datasheet
    let temp = 27.0 - (raw_temp as f32 * 3.3 / 4096.0 - 0.706) / 0.001721;
    let sign = if temp < 0.0 { -1.0 } else { 1.0 };
    let rounded_temp_x10: i16 = ((temp * 10.0) + 0.5 * sign) as i16;
    (rounded_temp_x10 as f32) / 10.0
}

pub struct Temperature<'a> {
    adc: Adc<'a, Async>,
    ts: Channel<'a>,
}

impl<'a> Temperature<'a> {
    /// Initialize the temperature sensor.
    pub fn new(adc_pin: ADC, adc_temp_pin: ADC_TEMP_SENSOR) -> Self {
        let adc = Adc::new(adc_pin, Irqs, Config::default());
        let ts = Channel::new_temp_sensor(adc_temp_pin);

        Self { adc, ts }
    }

    /// Read from the temperature sensor and return the value in degree celsius.
    pub async fn read(&mut self) -> Result<f32, Error> {
        let temp = self.adc.read(&mut self.ts).await?;
        let temp = convert_to_celsius(temp);

        Ok(temp)
    }
}
