use embassy_rp::{
    bind_interrupts,
    peripherals::USB,
    usb::{Driver, InterruptHandler},
};
use embassy_usb::{
    class::cdc_acm::{CdcAcmClass, State},
    Builder, Config, UsbDevice,
};
use static_cell::StaticCell;

bind_interrupts!(struct Irqs {
    USBCTRL_IRQ => InterruptHandler<USB>;
});

pub type SerialDevice = UsbDevice<'static, Driver<'static, USB>>;
pub type DeviceClass = CdcAcmClass<'static, Driver<'static, USB>>;

const DESCRIPTOR_SIZE: usize = 256;
type Descriptor = [u8; DESCRIPTOR_SIZE];

const BUFFER_SIZE: usize = 64;
type ControlBuffer = [u8; BUFFER_SIZE];

static DEVICE_DESCRIPTOR: StaticCell<Descriptor> = StaticCell::new();
static CONFIG_DESCRIPTOR: StaticCell<Descriptor> = StaticCell::new();
static BOS_DESCRIPTOR: StaticCell<Descriptor> = StaticCell::new();
static CONTROL_BUFFER: StaticCell<ControlBuffer> = StaticCell::new();
static STATE: StaticCell<State> = StaticCell::new();
static CLASS: StaticCell<DeviceClass> = StaticCell::new();

#[embassy_executor::task]
pub async fn usb_task(mut usb: SerialDevice) -> ! {
    usb.run().await
}

/// Initialize the USB serial interface and return the device and CdcAcm class (Abstract Control Model).
pub fn init_usb(usb: USB) -> (SerialDevice, &'static mut DeviceClass) {
    let driver = Driver::new(usb, Irqs);

    let mut config = Config::new(0xc0de, 0xcafe);
    config.manufacturer = Some("Embassy");
    config.product = Some("USB-serial");
    config.serial_number = Some("123456789");
    config.max_power = 100;
    config.max_packet_size_0 = 64;

    // Required for windows 7 compatibility.
    // https://developer.nordicsemi.com/nRF_Connect_SDK/doc/1.9.1/kconfig/CONFIG_CDC_ACM_IAD.html#help
    config.device_class = 0xEF;
    config.device_sub_class = 0x02;
    config.device_protocol = 0x01;
    config.composite_with_iads = true;

    let device_descriptor = DEVICE_DESCRIPTOR.init([0u8; DESCRIPTOR_SIZE]);
    let config_descriptor = CONFIG_DESCRIPTOR.init([0u8; DESCRIPTOR_SIZE]);
    let bos_descriptor = BOS_DESCRIPTOR.init([0u8; DESCRIPTOR_SIZE]);
    let control_buf = CONTROL_BUFFER.init([0u8; BUFFER_SIZE]);

    let state = STATE.init(State::new());
    let mut builder = Builder::new(
        driver,
        config,
        device_descriptor,
        config_descriptor,
        bos_descriptor,
        control_buf,
    );

    let class = CLASS.init(CdcAcmClass::new(&mut builder, state, 64));
    let usb = builder.build();

    (usb, &mut *class)
}
